from __future__ import annotations

from pathlib import Path

import click
from nltk.stem import SnowballStemmer
from stop_words import get_stop_words

from .format import BaseFormatter, JsonFormatter
from .preprocessing import (
    Corpus,
    Processor,
)
from .utils import try_read_file
from .wrappers import BM25, Tfidf

FORMATTER_CLASSES = {"plain": BaseFormatter, "json": JsonFormatter}
ALGORITHM_CLASSES = {"bm25": BM25, "tfidf": Tfidf}


@click.command()
@click.version_option()
@click.argument("reference-file", type=click.Path(), nargs=1, required=False)
@click.option(
    "-q",
    "--query",
    default=None,
    show_default=True,
    help="query option if no reference file is provided",
)
@click.option(
    "--directory",
    "-d",
    type=click.Path(),
    default=Path("./"),
    show_default="current directory",
    help="directory to scan for similar files",
    required=False,
)
@click.option(
    "--filename-pattern",
    "-f",
    type=str,
    default="*.*",
    help="filename pattern matching",
    show_default=True,
    required=False,
)
@click.option(
    "--recursive",
    "-R",
    is_flag=True,
    help="recursive search",
    required=False,
)
@click.option(
    "--algorithm",
    "-a",
    type=click.Choice(list(ALGORITHM_CLASSES.keys())),
    default="tfidf",
    show_default=True,
    help="text similarity algorithm",
    required=False,
)
@click.option(
    "--language",
    "-l",
    type=str,
    default="english",
    show_default=True,
    help="stemmer and stopwords language",
    required=False,
)
@click.option(
    "--min-chars",
    "-c",
    type=int,
    default=1,
    show_default=True,
    help="minimum document size (in number of characters) to be considered",
    required=False,
)
@click.option(
    "--absolute-paths",
    "-A",
    is_flag=True,
    help="show absolute rather than relative paths",
    required=False,
)
@click.option(
    "--max-results",
    "-m",
    type=int,
    default=10,
    show_default=True,
    help="maximum number of results",
    required=False,
)
@click.option(
    "--prefix",
    "-p",
    type=str,
    default="",
    show_default=True,
    help="result lines prefix",
    required=False,
)
@click.option(
    "--show-scores",
    "-s",
    is_flag=True,
    help="show similarity scores",
    required=False,
)
@click.option(
    "--hide-reference",
    "-h",
    is_flag=True,
    help="remove REFERENCE_FILE from results",
    required=False,
)
@click.option(
    "--heading",
    "-H",
    type=str,
    default="",
    show_default=True,
    help="results list heading",
    required=False,
)
@click.option(
    "--format",
    "-F",
    type=click.Choice(list(FORMATTER_CLASSES.keys())),
    default="plain",
    show_default=True,
    show_choices=True,
    help="output format",
    required=False,
)
@click.option(
    "--threshold",
    "-t",
    type=float,
    default=0.0,
    show_default=True,
    help="minimum score for a result to be shown",
)
def cli(
    reference_file,
    directory,
    filename_pattern,
    algorithm,
    max_results,
    language,
    min_chars,
    prefix,
    heading,
    show_scores,
    recursive,
    hide_reference,
    query,
    format,
    threshold,
    absolute_paths
):
    """'findlike' is a program that scans a given directory and returns the most
    similar documents in relation to REFERENCE_FILE or --query QUERY.

    Example using a reference file:

    $ findlike -d /path/to/my/notes my_recipe.md

    Example using a query:

    $ findlike -q "There is only one good, knowledge, and one evil, ignorance"
    """

    # Set up the reference text.
    if reference_file:
        reference_content = try_read_file(Path(reference_file))
    elif query:
        reference_content = query
    else:
        raise click.UsageError(
            "Neither REFERENCE_FILE nor --query QUERY was provided."
        )

    # Put together the list of documents to be analyzed.
    directory_path = Path(directory)
    glob_func = (
        directory_path.rglob
        if recursive
        else directory_path.glob
    )
    documents_paths = [x for x in glob_func(filename_pattern) if x.is_file()]

    # Create a corpus with the collected documents.
    corpus = Corpus(paths=documents_paths, min_chars=min_chars)

    # Set up the documents pre-processor.
    stemmer = SnowballStemmer(language).stem
    processor = Processor(
        stopwords=get_stop_words(language=language),
        stemmer=stemmer,
    )

    # Set up the similarity model.
    model = ALGORITHM_CLASSES[algorithm](processor=processor)
    model.fit(
        corpus.documents_ + [reference_content]
    )  # Add reference to avoid zero division
    scores = model.get_scores(source=reference_content)

    # Format and print results.
    formatter = FORMATTER_CLASSES[format](
        targets=corpus.paths_,
        scores=scores,
        max_results=max_results,
        show_scores=show_scores,
        hide_reference=hide_reference,
        prefix=prefix,
        heading=heading,
        threshold=threshold,
        absolute_paths=absolute_paths,
        is_query=bool(query)
        )
    formatted_results = formatter.format()
    print(formatted_results)



from __future__ import annotations

import re
from pathlib import Path
from typing import Callable

from .utils import try_read_file, compress

WORD_RE = re.compile(r"(?u)\b\w{2,}\b")
URL_RE = re.compile(r"\S*https?:\S*")

SCRIPT_PATH = Path(__file__).parent


class Processor:
    """Class containing preprocessing and tokenization rules.

    Args:
        junkchars (list): List of junk characters to be stripped from the text.
        stopwords (list): List of stopwords to be removed from the text.
        stemmer (nltk's stemmer): Stemmer provided by the nltk API.
    """

    def __init__(
        self,
        stopwords: list[str],
        stemmer: Callable,
    ):
        self.stopwords = stopwords
        self.stemmer = stemmer
        self._stopwords_re = re.compile(
            r"\b(" + r"|".join(stopwords) + r")\b\s*"
        )

    def preprocessor(self, text: str) -> str:
        """Remove fancy symbols and stopwords."""
        text = text.lower()
        text = text.translate({ord("’"): ord("'")})
        text = self._stopwords_re.sub("", text)
        text = URL_RE.sub("", text)
        return text

    def tokenizer(self, text: str) -> list[str]:
        """Run the tokenization and post-processing.
        This method should be called by the similarity algorithms.
        """
        tokens = self._tokenize(text)
        tokens = self._stemmize(tokens)
        return tokens

    def _tokenize(self, text: str) -> list[str]:
        """Preprocess a text and returns a list of tokens.
        This method should be called by the similarity algorithms.
        """
        words = WORD_RE.findall(text)
        return words

    def _stemmize(self, tokens: list[str]) -> list[str]:
        """Get only the stems from a list of words."""
        return [self.stemmer(w) for w in tokens]


class Corpus:
    """This wrapper provides easy access to a filtered corpus.

    Args:
        paths (list of Path): Document paths.
        min_chars (int): Minimum document size (in number of chars) to include
            in the corpus.
    Properties:
        documents_ (list of str): List of filtered document contents.
        paths_ (list of Path): List of filtered document paths.

    """

    def __init__(
        self,
        paths: list[Path],
        min_chars: int,
    ):
        self.paths = paths
        self.min_chars = min_chars

        self._loaded_documents: list[str | None]

        self.documents_: list[str]
        self.paths_: list[Path]

        self._load_documents()
        if min_chars:
            self._apply_min_chars_filter()
        self._prune_documents()
        self._prune_paths()

    def _load_documents(self):
        self._loaded_documents = [try_read_file(p) for p in self.paths]

    def _prune_paths(self):
        self.paths_ = compress(self.paths, self.documents_)

    def _prune_documents(self):
        self.documents_ = [x for x in self._loaded_documents if x]

    def _apply_min_chars_filter(self):
        """Apply min chars filter in both documents and documents paths"""
        self._loaded_documents = [
            doc if doc and len(doc) >= self.min_chars else None
            for doc in self._loaded_documents
        ]
        return self



import pytest
from nltk.stem import SnowballStemmer
from findlike.preprocessing import Processor
from stop_words import get_stop_words


@pytest.fixture
def processor():
    stemmer = SnowballStemmer("english").stem
    return Processor(
        stopwords=get_stop_words("english"), stemmer=stemmer
    )


def test_preprocessor(processor):
    text = "There’s a URL https://example.com in this text."
    processed_text = processor.preprocessor(text)
    assert processed_text == "'s url  text."


def test_tokenize(processor):
    text = "This is a test."
    tokens = processor._tokenize(text)
    assert tokens == ["This", "is", "test"]


def test_stemmize(processor):
    tokens = ["running", "cats", "quickly"]
    stemmed_tokens = processor._stemmize(tokens)
    assert stemmed_tokens == ["run", "cat", "quick"]


def test_tokenizer(processor):
    text = "There’s a word in this text. Running cats quickly!"
    tokens = processor.tokenizer(text)
    assert tokens == [
        "there",
        "word",
        "in",
        "this",
        "text",
        "run",
        "cat",
        "quick",
    ]


def test_empty_text(processor):
    text = ""
    tokens = processor.tokenizer(text)
    assert tokens == []


def test_text_with_only_stopwords(processor):
    text = "This is a the an"
    tokens = processor.preprocessor(text)
    assert tokens == ""


def test_text_with_special_characters(processor):
    text = "#This is a test! @user #hashtag"
    tokens = processor.tokenizer(processor.preprocessor(text))
    assert tokens == ["test", "user", "hashtag"]


def test_text_with_numbers(processor):
    text = "There are 100 cats"
    tokens = processor.tokenizer(processor.preprocessor(text))
    assert tokens == ["100", "cat"]


def test_text_with_unicode_characters(processor):
    text = "café naïve"
    processed_text = processor.preprocessor(text)
    tokens = processor.tokenizer(processed_text)
    assert tokens == ["café", "naïv"]



from __future__ import annotations

import json
from pathlib import Path


class BaseFormatter:
    """Format results in an org-compatible format with links.

    Args:
        input_filename (Path): path to the filename that will be used as
            reference.
        target_filenames (Path): Glob containing the path to the documents
            whose similarity with the input filename will be estimated.
        scores (array-like): List of similarity scores with the same number of
            documents in target_filenames plus one (accounting for the
            input_filename).
        num_results (int): How many similar entries to list at the end of the buffer.
        show_scores (bool): Whether to prepend the results with the similarity score.
        remove_first (bool): Remove first result from the scores list. Useful if
            the source document is inside the same directory as the target documents,
            and you don't want to see it included in the list for obvious reasons.
            Default is False.
        prefix (str): Prefix to add to each item in the printed list.
        heading (str): List heading (title).
        hide_reference (bool): Omit the reference file from the final results.
        absolute_paths (bool): Show absolute paths instead of relative ones.
        is_query (bool): Flag whether the formatted results come from a query search.

    Returns:
        Formatted output with the list of most similar documents, sorted in descending
        order of similarity.
    """

    def __init__(
        self,
        targets: list[Path],
        scores: list[float],
        max_results: int = 10,
        show_scores: bool = False,
        hide_reference: bool = False,
        prefix: str = "",
        heading: str = "",
        threshold: float = 0.0,
        absolute_paths: bool = False,
        is_query: bool = False,
    ):
        self.targets = targets
        self.scores = scores
        self.max_results = max_results
        self.show_scores = show_scores
        self.hide_reference = hide_reference
        self.prefix = prefix
        self.heading = heading
        self.threshold = threshold
        self.absolute_paths = absolute_paths
        self.is_query = is_query

        self._scores_targets: list[tuple[float, Path]]
        self._format_targets()._zip_pairs()._filter_pairs()

    def _zip_pairs(self):
        self._scores_targets = list(zip(self.scores, self.targets))
        return self

    def _filter_pairs(self):
        """Remove reference doc if needed and limit list to max number of results."""

        # Sort pairs according to distance (descending).
        self._scores_targets = sorted(
            self._scores_targets, key=lambda x: x[0], reverse=True
        )

        # Remove entries whose score are below the threshold.
        self._scores_targets = [
            x for x in self._scores_targets if x[0] >= self.threshold
        ]

        # Limit number of results, disregarding the first entry if `remove_reference`
        # flag was passed.
        start_pos = int(self.hide_reference and not self.is_query)
        range = slice(start_pos, start_pos + self.max_results)
        self._scores_targets = self._scores_targets[range]

        return self

    def _format_targets(self):
        if self.absolute_paths:
            self.targets = [x.resolve() for x in self.targets]
        return self

    def _format_score(self, score):
        return f"{score:.2f}" + " " if self.show_scores else ""

    def format(self):
        entries = []
        if self.heading:
            entries.append(self.heading)
        entries += [
            f"{self.prefix}{self._format_score(score)}{target}"
            for score, target in self._scores_targets
        ]
        output = "\n".join(entries)
        return output


class JsonFormatter(BaseFormatter):
    def format(self):
        if self.heading:
            print(self.heading)
        if self.show_scores:
            entries = [
                {"score": score, "target": str(target)}
                for score, target in self._scores_targets
            ]
        else:
            entries = [
                {"target": str(target)} for _, target in self._scores_targets
            ]
        json_data = json.dumps(entries)
        return json_data



from __future__ import annotations

from typing import Iterable, Any
from pathlib import Path


def try_read_file(filename: Path) -> str | None:
    with filename.open() as f:
        try:
            document = f.read()
        except UnicodeDecodeError:
            document = None
    return document


def compress(data: Iterable[Any], selectors: Iterable[Any]) -> list[Any]:
    return [d for d, s in zip(data, selectors) if s]



import pytest
from findlike.format import BaseFormatter, JsonFormatter
from pathlib import Path
import json

# Sample test data
targets = [Path("file1.txt"), Path("file2.txt"), Path("file3.txt")]
scores = [0.8, 0.6, 0.9]


class TestBaseFormatter:
    @pytest.mark.parametrize("max_results, expected_length", [(2, 2), (5, 3)])
    def test_max_results(self, max_results, expected_length):
        formatter = BaseFormatter(targets, scores, max_results=max_results)
        output = formatter.format()
        assert len(output.split("\n")) == expected_length

    def test_show_scores(self):
        formatter = BaseFormatter(targets, scores, show_scores=True)
        output = formatter.format()
        assert all(
            any(char.isdigit() for char in line) for line in output.split("\n")
        )

    def test_hide_reference(self):
        formatter = BaseFormatter(targets, scores, hide_reference=True)
        output = formatter.format()
        assert len(output.split("\n")) == 2

    def test_prefix(self):
        formatter = BaseFormatter(targets, scores, prefix="PREFIX_")
        output = formatter.format()
        assert all(line.startswith("PREFIX_") for line in output.split("\n"))

    def test_heading(self):
        heading = "Sample Heading"
        formatter = BaseFormatter(targets, scores, heading=heading)
        output = formatter.format()
        assert output.split("\n")[0] == heading

    def test_threshold(self):
        formatter = BaseFormatter(targets, scores, threshold=0.7)
        output = formatter.format()
        assert len(output.split("\n")) == 2

    def test_absolute_paths(self):
        formatter = BaseFormatter(targets, scores, absolute_paths=True)
        output = formatter.format()
        assert all(Path(line).is_absolute() for line in output.split("\n"))


class TestJsonFormatter:
    def test_valid_json(self):
        formatter = JsonFormatter(targets, scores)
        output = formatter.format()
        try:
            data = json.loads(output)
        except json.JSONDecodeError:
            pytest.fail("Output is not valid JSON")

    @pytest.mark.parametrize(
        "show_scores, expected_keys",
        [(True, ["score", "target"]), (False, ["target"])],
    )
    def test_keys_in_json(self, show_scores, expected_keys):
        formatter = JsonFormatter(targets, scores, show_scores=show_scores)
        output = formatter.format()
        data = json.loads(output)
        for entry in data:
            assert set(entry.keys()) == set(expected_keys)

    def test_hide_reference(self):
        formatter = JsonFormatter(targets, scores, hide_reference=True)
        output = formatter.format()
        data = json.loads(output)
        assert len(data) == 2

    def test_threshold(self):
        formatter = JsonFormatter(targets, scores, threshold=0.7)
        output = formatter.format()
        data = json.loads(output)
        assert len(data) == 2



from __future__ import annotations

from rank_bm25 import BM25Okapi
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity


class Tfidf:
    """Scikit-learn's TF-IDF wrapper.

    Args:
        processor (Processor): Processor object.
    """

    def __init__(self, processor, **kwargs):
        self.processor = processor

        self._vectorizer = TfidfVectorizer(
            tokenizer=self.processor.tokenizer,
            preprocessor=self.processor.preprocessor,
            token_pattern=None,
            **kwargs,
        )

    def fit(self, documents: list[str]):
        self._vectorizer.fit(documents)
        self.documents_embeddings_ = self._vectorizer.transform(documents)

    def get_scores(self, source: str):
        self.source_embeddings_ = self._vectorizer.transform([source])
        scores = cosine_similarity(
            self.source_embeddings_, self.documents_embeddings_
        ).flatten()
        return scores


class BM25:
    """Okapi BM25 wrapper.

    Args:
        processor (Processor): Processor object.
    """

    def __init__(self, processor):
        self.processor = processor

    def fit(self, documents: list[str]):
        """Fit IDF to documents X"""
        clean_docs = [self.processor.preprocessor(d) for d in documents]
        tokenized_docs = [self.processor.tokenizer(d) for d in clean_docs]
        self._model = BM25Okapi(tokenized_docs)

    def get_scores(self, source: str):
        clean_source = self.processor.preprocessor(source)
        tokenized_source = self.processor.tokenizer(clean_source)
        scores = self._model.get_scores(tokenized_source)
        return scores
